#ifndef TYPON_FUNDAMENTAL_OPTIONAL_HPP_INCLUDED
#define TYPON_FUNDAMENTAL_OPTIONAL_HPP_INCLUDED

#include <type_traits>


namespace typon::fdt
{

  template <typename T, unsigned char N>
    requires std::is_trivially_copyable_v<T>
  struct optional
  {
    static_assert(N > 0, "N must be greater than 0");

    template <unsigned char I>
      requires (I <= N)
    using state = std::integral_constant<unsigned char, I>;

    unsigned char _state;
    union
    {
      T _value;
    };

    optional() noexcept : _state(0) {}

    template <unsigned char I>
      requires (I < N)
    optional(state<I> state) noexcept : _state(state) {}

    optional(T value) noexcept : _state(N), _value(value) {}

    ~optional()
    {
      if (_state == N)
      {
        std::destroy_at(std::addressof(_value));
      }
    }

    operator bool() noexcept
    {
      return _state == N;
    }

    template <unsigned char I>
    bool match(state<I> state) noexcept
    {
      return state() == _state;
    }

    T * operator->() noexcept
    {
      return std::addressof(_value);
    }

    T & operator*() noexcept
    {
      return _value;
    }
  };

}


#endif // TYPON_FUNDAMENTAL_OPTIONAL_HPP_INCLUDED
